@@ -1,3 +1,5 @@
+#include <sys/select.h>
+#include <sys/time.h>
 #include <sys/types.h>
 #include <sys/wait.h>
 
@@ -47,8 +49,8 @@ struct row {
 };
 
 struct pane {
-	int x; /* absolute x position of the window on the screen */
-	int y; /* absolute y position of the window on the screen */
+	int x; /* absolute x position on the screen */
+	int y; /* absolute y position on the screen */
 	int width; /* absolute width of the window */
 	int height; /* absolute height of the window */
 	off_t pos; /* focused row position */
@@ -66,8 +68,8 @@ struct pane {
 struct scrollbar {
 	int tickpos;
 	int ticksize;
-	int x; /* absolute x position of the window on the screen */
-	int y; /* absolute y position of the window on the screen */
+	int x; /* absolute x position on the screen */
+	int y; /* absolute y position on the screen */
 	int size; /* absolute size of the bar */
 	int focused; /* has focus or not */
 	int hidden; /* is visible or not */
@@ -75,8 +77,8 @@ struct scrollbar {
 };
 
 struct statusbar {
-	int x; /* absolute x position of the window on the screen */
-	int y; /* absolute y position of the window on the screen */
+	int x; /* absolute x position on the screen */
+	int y; /* absolute y position on the screen */
 	int width; /* absolute width of the bar */
 	char *text; /* data */
 	int hidden; /* is visible or not */
@@ -88,19 +90,20 @@ struct statusbar {
 /* feed info */
 struct feed {
 	char         *name;     /* feed name */
-	char         *path;     /* path to feed or NULL from stdin */
+	char         *path;     /* path to feed or NULL for stdin */
 	unsigned long totalnew; /* amount of new items per feed */
 	unsigned long total;    /* total items */
 	FILE *fp;               /* file pointer */
 };
 
 struct item {
+	char *link; /* separate link field (always loaded in case of urlfile) */
 	char *fields[FieldLast];
 	char *line; /* allocated split line */
 	time_t timestamp;
 	int timeok;
 	int isnew;
-	off_t offset; /* line offset in file */
+	off_t offset; /* line offset in file for lazyload */
 };
 
 #undef err
@@ -109,7 +112,10 @@ void err(int, const char *, ...);
 void alldirty(void);
 void cleanup(void);
 void draw(void);
+int isurlnew(const char *);
+void markread(struct pane *, off_t, off_t, int);
 void pane_draw(struct pane *);
+void readurls(void);
 void sighandler(int);
 void updategeom(void);
 void updatesidebar(int);
@@ -131,13 +137,17 @@ static struct feed *feeds;
 static struct feed *curfeed;
 static size_t nfeeds; /* amount of feeds */
 static time_t comparetime;
+static char *urlfile, **urls;
+static size_t nurls;
+
+volatile sig_atomic_t sigstate = 0;
 
 /* config */
 
 /* Allow to lazyload items when a file is specified? This saves memory but
    increases some latency when seeking items. It also causes issues if the
    feed is changed while having the UI open (and offsets are changed). */
-static int lazyload = 0;
+/*#define LAZYLOAD 1 */
 
 static char *plumber = "xdg-open"; /* environment variable: $SFEED_PLUMBER */
 static char *piper = "less"; /* environment variable: $SFEED_PIPER */
@@ -428,8 +438,8 @@ cleanup(void)
 	/* restore terminal settings */
 	tcsetattr(0, TCSANOW, &tsave);
 
-	appmode(0);
 	cursormode(1);
+	appmode(0);
 	clearscreen();
 
 	/* xterm mouse-mode */
@@ -483,8 +493,8 @@ init(void)
 
 	resizewin();
 
-	cursormode(0);
 	appmode(1);
+	cursormode(0);
 
 	/* xterm mouse-mode */
 	if (usemouse)
@@ -496,6 +506,7 @@ init(void)
 	sigemptyset(&sa.sa_mask);
 	sa.sa_flags = SA_RESTART; /* require BSD signal semantics */
 	sa.sa_handler = sighandler;
+	sigaction(SIGHUP, &sa, NULL);
 	sigaction(SIGINT, &sa, NULL);
 	sigaction(SIGTERM, &sa, NULL);
 	sigaction(SIGWINCH, &sa, NULL);
@@ -510,14 +521,14 @@ void
 pipeitem(const char *cmd, struct item *item, int wantoutput)
 {
 	FILE *fp;
-	int i, pid, wpid;
+	int i, pid, wpid, status;
 
 	if (wantoutput)
 		cleanup();
+
 	switch ((pid = fork())) {
 	case -1:
 		err(1, "fork");
-		return;
 	case 0:
 		if (!wantoutput) {
 			dup2(devnullfd, 1);
@@ -525,28 +536,26 @@ pipeitem(const char *cmd, struct item *item, int wantoutput)
 		}
 
 		errno = 0;
-		if (!(fp = popen(cmd, "w"))) {
-			fputs("popen: ", stderr);
-			perror(NULL);
-			_exit(1);
-		}
+		if (!(fp = popen(cmd, "w")))
+			err(1, "popen");
 		for (i = 0; i < FieldLast; i++) {
 			if (i)
 				fputc('\t', fp);
 			fputs(item->fields[i], fp);
 		}
 		fputc('\n', fp);
-		_exit(pclose(fp));
-		break;
+		status = pclose(fp);
+		status = WIFEXITED(status) ? WEXITSTATUS(status) : 127;
+		_exit(status);
 	default:
 		while ((wpid = wait(NULL)) >= 0 && wpid != pid)
 			;
-	}
 
-	if (wantoutput) {
-		updatesidebar(onlynew);
-		updatetitle();
-		init();
+		if (wantoutput) {
+			updatesidebar(onlynew);
+			updatetitle();
+			init();
+		}
 	}
 }
 
@@ -556,7 +565,6 @@ plumb(const char *cmd, char *url)
 	switch (fork()) {
 	case -1:
 		err(1, "fork");
-		return;
 	case 0:
 		dup2(devnullfd, 1);
 		dup2(devnullfd, 2);
@@ -568,7 +576,7 @@ plumb(const char *cmd, char *url)
 struct row *
 pane_row_get(struct pane *p, off_t pos)
 {
-	if (pos >= p->nrows)
+	if (pos < 0 || pos >= p->nrows)
 		return NULL;
 
 	if (p->row_get)
@@ -627,18 +635,28 @@ pane_row_draw(struct pane *p, off_t pos)
 }
 
 void
-pane_setpos(struct pane *p, off_t pos)
+pane_row_redraw(struct pane *p, off_t pos)
 {
 	off_t prev;
 
-	if (pos == p->pos)
-		return; /* no change */
-	if (!p->nrows)
-		return; /* invalid */
+	prev = p->pos;
+	p->pos = pos;
+	pane_row_draw(p, prev); /* draw previous row again */
+	pane_row_draw(p, p->pos); /* draw new highlighted row */
+	fflush(stdout); /* flush and update directly */
+}
+
+void
+pane_setpos(struct pane *p, off_t pos)
+{
 	if (pos < 0)
 		pos = 0; /* clamp */
+	if (!p->nrows)
+		return; /* invalid */
 	if (pos >= p->nrows)
 		pos = p->nrows - 1; /* clamp */
+	if (pos == p->pos)
+		return; /* no change */
 
 	/* is on different scroll region? mark dirty */
 	if (((p->pos - (p->pos % p->height)) / p->height) !=
@@ -648,11 +666,7 @@ pane_setpos(struct pane *p, off_t pos)
 		pane_draw(p);
 	} else {
 		/* only redraw the 1 or 2 dirty rows */
-		prev = p->pos;
-		p->pos = pos;
-		pane_row_draw(p, prev); /* draw previous row again */
-		pane_row_draw(p, p->pos); /* draw new highlighted row */
-		fflush(stdout); /* flush and update directly */
+		pane_row_redraw(p, pos);
 	}
 }
 
@@ -862,15 +876,29 @@ int
 readch(void)
 {
 	unsigned char b;
-	ssize_t n;
+	fd_set readfds;
+	struct timeval tv;
 
-	n = read(0, &b, 1);
-	if (n == 0)
-		return EOF;
-	else if (n == -1)
-		err(1, "read");
+	for (;;) {
+		FD_ZERO(&readfds);
+		FD_SET(0, &readfds);
+		tv.tv_sec = 0;
+		tv.tv_usec = 250000; /* 250ms */
+		switch (select(1, &readfds, NULL, NULL, &tv)) {
+		case -1:
+			if (errno != EINTR)
+				err(1, "select");
+			return -2; /* EINTR: like a signal */
+		case 0:
+			return -3; /* time-out */
+		}
 
-	return (int)b;
+		switch (read(0, &b, 1)) {
+		case -1: err(1, "read");
+		case 0: return EOF;
+		default: return (int)b;
+		}
+	}
 }
 
 char *
@@ -903,6 +931,8 @@ lineeditor(void)
 			continue;
 		} else if (ch >= ' ') {
 			write(1, &ch, 1);
+		} else if (ch < 0) {
+			continue; /* process signals later */
 		}
 		input[nchars++] = ch;
 	}
@@ -972,7 +1002,13 @@ linetoitem(char *line, struct item *item)
 	char *fields[FieldLast];
 	time_t parsedtime;
 
+	item->line = line;
 	parseline(line, fields);
+	memcpy(item->fields, fields, sizeof(fields));
+	if (urlfile)
+		item->link = estrdup(fields[FieldLink]);
+	else
+		item->link = NULL;
 
 	parsedtime = 0;
 	if (!strtotime(fields[FieldUnixTimestamp], &parsedtime)) {
@@ -983,10 +1019,6 @@ linetoitem(char *line, struct item *item)
 		item->timeok = 0;
 	}
 
-	item->line = line;
-	item->isnew = (parsedtime >= comparetime);
-	memcpy(item->fields, fields, sizeof(fields));
-
 	return 0;
 }
 
@@ -1019,7 +1051,8 @@ feed_getitems(struct feed *f, FILE *fp, struct item **items, size_t *nitems)
 			if (line[linelen - 1] == '\n')
 				line[--linelen] = '\0';
 
-			if (lazyload && f->path) {
+#ifdef LAZYLOAD
+			if (f->path) {
 				linetoitem(line, item);
 
 				/* data is ignored here, will be lazy-loaded later. */
@@ -1028,6 +1061,9 @@ feed_getitems(struct feed *f, FILE *fp, struct item **items, size_t *nitems)
 			} else {
 				linetoitem(estrdup(line), item);
 			}
+#else
+			linetoitem(estrdup(line), item);
+#endif
 
 			(*nitems)++;
 		}
@@ -1043,6 +1079,29 @@ err:
 	return ret;
 }
 
+void
+updatenewitems(struct feed *f)
+{
+	struct pane *p;
+	struct row *row;
+	struct item *item;
+	size_t i;
+
+	p = &panes[PaneItems];
+	f->totalnew = 0;
+	for (i = 0; i < p->nrows; i++) {
+		row = &(p->rows[i]); /* do not use pane_row_get */
+		item = (struct item *)row->data;
+		if (urlfile)
+			item->isnew = isurlnew(item->link);
+		else
+			item->isnew = (item->timeok && item->timestamp >= comparetime);
+		row->bold = item->isnew;
+		f->totalnew += item->isnew;
+	}
+	f->total = p->nrows;
+}
+
 void
 feed_load(struct feed *f, FILE *fp)
 {
@@ -1052,8 +1111,10 @@ feed_load(struct feed *f, FILE *fp)
 	struct row *row;
 	size_t i;
 
-	for (i = 0; i < nitems; i++)
+	for (i = 0; i < nitems; i++) {
 		free(items[i].line);
+		free(items[i].link);
+	}
 	free(items);
 	items = NULL;
 	nitems = 0;
@@ -1064,24 +1125,20 @@ feed_load(struct feed *f, FILE *fp)
 	p->nrows = 0;
 
 	if (feed_getitems(f, fp, &items, &nitems) == -1)
-		err(1, "%s: %s", __func__, f->path);
-
-	f->totalnew = 0;
-	f->total = nitems;
-	for (i = 0; i < nitems; i++)
-		f->totalnew += items[i].isnew;
+		err(1, "%s: %s", __func__, f->name);
 
 	p->pos = 0;
 	p->nrows = nitems;
 	p->rows = ecalloc(sizeof(p->rows[0]), nitems + 1);
 	for (i = 0; i < nitems; i++) {
-		row = &(p->rows[i]);
+		row = &(p->rows[i]); /* do not use pane_row_get */
 		row->text = ""; /* custom formatter */
-		row->bold = items[i].isnew;
 		row->data = &items[i];
 	}
+
+	updatenewitems(f);
+
 	p->dirty = 1;
-	scrollbars[PaneItems].dirty = 1;
 }
 
 void
@@ -1093,14 +1150,19 @@ feed_count(struct feed *f, FILE *fp)
 	ssize_t linelen;
 	time_t parsedtime;
 
+	f->totalnew = f->total = 0;
 	while ((linelen = getline(&line, &linesize, fp)) > 0) {
 		if (line[linelen - 1] == '\n')
 			line[--linelen] = '\0';
 		parseline(line, fields);
 
-		parsedtime = 0;
-		if (!strtotime(fields[FieldUnixTimestamp], &parsedtime))
-			f->totalnew += (parsedtime >= comparetime);
+		if (urlfile) {
+			f->totalnew += isurlnew(fields[FieldLink]);
+		} else {
+			parsedtime = 0;
+			if (!strtotime(fields[FieldUnixTimestamp], &parsedtime))
+				f->totalnew += (parsedtime >= comparetime);
+		}
 		f->total++;
 	}
 	free(line);
@@ -1149,7 +1211,6 @@ feeds_load(struct feed *feeds, size_t nfeeds)
 
 	for (i = 0; i < nfeeds; i++) {
 		f = &feeds[i];
-		f->totalnew = f->total = 0;
 
 		if (f->path) {
 			if (f->fp) {
@@ -1160,27 +1221,48 @@ feeds_load(struct feed *feeds, size_t nfeeds)
 					err(1, "fopen: %s", f->path);
 			}
 		}
+		if (!f->fp) {
+			/* reading from stdin, just recount new */
+			if (f == curfeed)
+				updatenewitems(f);
+			continue;
+		}
 
 		/* load first items, because of first selection or stdin. */
-		if (i == 0)
+		if (f == curfeed) {
 			feed_load(f, f->fp);
-		else
+		} else {
 			feed_count(f, f->fp);
-
-		if (f->path && f->fp) {
-			fclose(f->fp);
-			f->fp = NULL;
+			if (f->path && f->fp) {
+				fclose(f->fp);
+				f->fp = NULL;
+			}
 		}
 	}
 }
 
+void
+feeds_reloadall(void)
+{
+	off_t pos;
+
+	pos = panes[PaneItems].pos; /* store numeric position */
+	readurls();
+	feeds_set(curfeed); /* close and reopen feed if possible */
+	feeds_load(feeds, nfeeds);
+	/* restore numeric position */
+	pane_setpos(&panes[PaneItems], pos);
+	updatesidebar(onlynew);
+	updatetitle();
+}
+
 void
 updatesidebar(int onlynew)
 {
 	struct pane *p;
 	struct row *row;
 	struct feed *feed;
-	size_t i;
+	size_t i, nrows;
 	int len, width;
 
 	p = &panes[PaneFeeds];
@@ -1188,13 +1270,12 @@ updatesidebar(int onlynew)
 	if (!p->rows)
 		p->rows = ecalloc(sizeof(p->rows[0]), nfeeds + 1);
 
-	p->nrows = 0;
-	p->width = 0;
+	nrows = 0;
 	width = 0;
 	for (i = 0; i < nfeeds; i++) {
 		feed = &feeds[i];
 
-		row = &(p->rows[p->nrows]);
+		row = &(p->rows[nrows]);
 		row->text = ""; /* custom formatter is used */
 		row->bold = (feed->totalnew > 0);
 		row->data = feed;
@@ -1206,25 +1287,32 @@ updatesidebar(int onlynew)
 		if (onlynew && feed->totalnew == 0)
 			continue;
 
-		p->nrows++;
+		nrows++;
 	}
+
+	if (p->width != width) {
+		for (i = 0; i < PaneLast; i++)
+			panes[i].dirty = 1;
+	} else {
+		p->dirty = 1;
+	}
+	p->nrows = nrows;
 	p->width = width;
-	p->dirty = 1;
+	if (!p->nrows)
+		p->pos = 0;
+	else if (p->pos >= p->nrows)
+		p->pos = p->nrows - 1;
 }
 
 void
 sighandler(int signo)
 {
 	switch (signo) {
+	case SIGHUP:
 	case SIGINT:
 	case SIGTERM:
-		cleanup();
-		_exit(128 + signo);
-		break;
 	case SIGWINCH:
-		resizewin();
-		updategeom();
-		draw();
+		sigstate = signo;
 		break;
 	}
 }
@@ -1317,14 +1405,16 @@ mousereport(int button, int release, int x, int y)
 				row = pane_row_get(p, pos);
 				f = (struct feed *)row->data;
 				feeds_set(f);
-				feed_load(f, f->fp);
+				if (f->fp)
+					feed_load(f, f->fp);
 				/* redraw row: counts could be changed */
-				pane_row_draw(p, pos);
+				updatesidebar(onlynew);
 				updatetitle();
 			} else if (i == PaneItems) {
 				if (dblclick && !changedpane) {
 					row = pane_row_get(&panes[PaneItems], pos);
 					item = (struct item *)row->data;
+					markread(p, p->pos, p->pos, 1);
 					plumb(plumber, item->fields[FieldLink]);
 				}
 			}
@@ -1337,6 +1427,7 @@ mousereport(int button, int release, int x, int y)
 				p = &panes[PaneItems];
 				row = pane_row_get(p, p->pos);
 				item = (struct item *)row->data;
+				markread(p, p->pos, p->pos, 1);
 				pipeitem(piper, item, 1);
 			}
 			break;
@@ -1396,7 +1487,7 @@ item_row_get(struct pane *p, off_t pos)
 	item = (struct item *)itemrow->data;
 
 	f = curfeed;
-	if (f && f->path && !item->line) {
+	if (f && f->path && f->fp && !item->line) {
 		if (fseek(f->fp, item->offset, SEEK_SET))
 			err(1, "fseek: %s", f->path);
 		linelen = getline(&line, &linesize, f->fp);
@@ -1440,6 +1531,120 @@ item_row_format(struct pane *p, struct row *row)
 	return text;
 }
 
+void
+markread(struct pane *p, off_t from, off_t to, int isread)
+{
+	struct row *row;
+	struct item *item;
+	FILE *fp;
+	off_t i;
+	const char *cmd;
+	int changed, isnew = !isread, pid, wpid, status;
+
+	if (!urlfile || !p->nrows)
+		return;
+
+	if (isread) {
+		if (!(cmd = getenv("SFEED_MARK_READ")))
+			cmd = "sfeed_markread read";
+	} else {
+		if (!(cmd = getenv("SFEED_MARK_UNREAD")))
+			cmd = "sfeed_markread unread";
+	}
+
+	switch ((pid = fork())) {
+	case -1:
+		err(1, "fork");
+	case 0:
+		dup2(devnullfd, 1);
+		dup2(devnullfd, 2);
+
+		errno = 0;
+		if (!(fp = popen(cmd, "w")))
+			err(1, "popen");
+
+		for (i = from; i <= to; i++) {
+			row = &(p->rows[i]);
+			item = (struct item *)row->data;
+			if (item->isnew != isnew) {
+				fputs(item->link, fp);
+				fputc('\n', fp);
+			}
+		}
+		status = pclose(fp);
+		status = WIFEXITED(status) ? WEXITSTATUS(status) : 127;
+		_exit(status);
+	default:
+		while ((wpid = wait(&status)) >= 0 && wpid != pid)
+			;
+
+		/* fail: exit statuscode was non-zero */
+		if (status)
+			break;
+		for (i = from, changed = 0; i <= to && i < p->nrows; i++) {
+			row = &(p->rows[i]);
+			item = (struct item *)row->data;
+			if (item->isnew != isnew) {
+				row->bold = item->isnew = isnew;
+				curfeed->totalnew += isnew ? 1 : -1;
+				changed++;
+			}
+		}
+		if (changed) {
+			if (from != to || from != p->pos)
+				p->dirty = 1;
+			else
+				pane_row_redraw(p, from);
+		}
+		updatesidebar(onlynew);
+		updatetitle();
+	}
+}
+
+int
+urlcmp(const void *v1, const void *v2)
+{
+	return strcmp(*((char **)v1), *((char **)v2));
+}
+
+void
+readurls(void)
+{
+	FILE *fp;
+	char *line = NULL;
+	size_t linesiz = 0, cap = 0;
+	ssize_t n;
+
+	while (nurls > 0)
+		free(urls[--nurls]);
+	free(urls);
+	urls = NULL;
+	nurls = 0;
+
+	if (!urlfile || !(fp = fopen(urlfile, "rb")))
+		return;
+
+	while ((n = getline(&line, &linesiz, fp)) > 0) {
+		if (line[n - 1] == '\n')
+			line[--n] = '\0';
+		if (nurls + 1 >= cap) {
+			cap = cap ? cap * 2 : 16;
+			urls = erealloc(urls, cap * sizeof(char *));
+		}
+		urls[nurls++] = estrdup(line);
+	}
+	fclose(fp);
+	free(line);
+
+	qsort(urls, nurls, sizeof(char *), urlcmp);
+}
+
+int
+isurlnew(const char *url)
+{
+	return bsearch(&url, urls, nurls, sizeof(char *), urlcmp) == NULL;
+}
+
 int
 main(int argc, char *argv[])
 {
@@ -1459,6 +1664,12 @@ main(int argc, char *argv[])
 		plumber = tmp;
 	if ((tmp = getenv("SFEED_PIPER")))
 		piper = tmp;
+	urlfile = getenv("SFEED_URL_FILE");
+
+	panes[PaneFeeds].row_format = feed_row_format;
+	panes[PaneFeeds].row_match = feed_row_match;
+	panes[PaneItems].row_format = item_row_format;
+	panes[PaneItems].row_get = item_row_get;
 
 	feeds = ecalloc(argc, sizeof(struct feed));
 	if (argc == 1) {
@@ -1476,9 +1687,9 @@ main(int argc, char *argv[])
 		}
 		nfeeds = argc - 1;
 	}
-	feeds_set(NULL);
-	feeds_load(feeds, nfeeds);
+	readurls();
 	feeds_set(&feeds[0]);
+	feeds_load(feeds, nfeeds);
 
 	if (!isatty(0)) {
 		if ((fd = open("/dev/tty", O_RDONLY)) == -1)
@@ -1489,11 +1700,6 @@ main(int argc, char *argv[])
 	if (argc == 1)
 		feeds[0].fp = NULL;
 
-	panes[PaneFeeds].row_format = feed_row_format;
-	panes[PaneFeeds].row_match = feed_row_match;
-	panes[PaneItems].row_format = item_row_format;
-	panes[PaneItems].row_get = item_row_get;
-
 	if (argc > 1) {
 		panes[PaneFeeds].hidden = 0;
 		selpane = PaneFeeds;
@@ -1510,17 +1716,21 @@ main(int argc, char *argv[])
 	init();
 	draw();
 
-	while ((ch = readch()) != EOF) {
+	while (1) {
+		if ((ch = readch()) < 0)
+			goto event;
 		switch (ch) {
 		case '\x1b':
-			ch = readch();
+			if ((ch = readch()) < 0)
+				goto event;
 			if (ch != '[' && ch != 'O')
 				continue; /* unhandled */
-			switch ((ch = readch())) {
-			case EOF: goto end;
+			if ((ch = readch()) < 0)
+				goto event;
+			switch (ch) {
 			case 'M': /* reported mouse event */
-				if ((ch = readch()) == EOF)
-					goto end;
+				if ((ch = readch()) < 0)
+					goto event;
 				/* button numbers (0 - 2) encoded in lowest 2 bits
 				   release does not indicate which button (so set to 0).
 				   Handle extended buttons like scrollwheels
@@ -1536,10 +1746,12 @@ main(int argc, char *argv[])
 				}
 
 				/* X10 mouse-encoding */
-				if ((x = readch()) == EOF)
-					goto end;
-				if ((y = readch()) == EOF)
-					goto end;
+				if ((ch = readch()) < 0)
+					goto event;
+				x = ch;
+				if ((ch = readch()) < 0)
+					goto event;
+				y = ch;
 				mousereport(button, release, x - 33, y - 33);
 				break;
 			case 'A': goto keyup;    /* arrow up */
@@ -1549,15 +1761,21 @@ main(int argc, char *argv[])
 			case 'F': goto endpos;   /* end */
 			case 'H': goto startpos; /* home */
 			case '4': /* end */
-				if ((ch = readch()) == '~')
+				if ((ch = readch()) < 0)
+					goto event;
+				if (ch == '~')
 					goto endpos;
 				continue;
 			case '5': /* page up */
-				if ((ch = readch()) == '~')
+				if ((ch = readch()) < 0)
+					goto event;
+				if (ch == '~')
 					goto prevpage;
 				continue;
 			case '6': /* page down */
-				if ((ch = readch()) == '~')
+				if ((ch = readch()) < 0)
+					goto event;
+				if (ch == '~')
 					goto nextpage;
 				continue;
 			}
@@ -1635,20 +1853,11 @@ nextpage:
 				}
 			}
 			break;
-		case 12: /* ^L */
-		case 'r': /* update window dimensions and redraw */
+		case 12: /* ^L, redraw */
 			alldirty();
 			break;
 		case 'R': /* reload all files */
-			if (nfeeds == 1 && !feeds[0].path)
-				break; /* do not reload when read from stdin */
-			feeds_set(NULL);
-			feeds_load(feeds, nfeeds);
-			feeds_set(&feeds[0]);
-			panes[PaneFeeds].pos = 0;
-			updatesidebar(onlynew);
-			updategeom();
-			updatetitle();
+			feeds_reloadall();
 			break;
 		case 'a': /* attachment */
 		case 'e': /* enclosure */
@@ -1663,7 +1872,6 @@ nextpage:
 		case 'm': /* toggle mouse mode */
 			usemouse = !usemouse;
 			mousemode(usemouse);
-			fflush(stdout);
 			break;
 		case 's': /* toggle sidebar */
 			panes[PaneFeeds].hidden = !panes[PaneFeeds].hidden;
@@ -1683,14 +1891,16 @@ nextpage:
 				row = pane_row_get(p, p->pos);
 				f = (struct feed *)row->data;
 				feeds_set(f);
-				feed_load(f, f->fp);
+				if (f->fp)
+					feed_load(f, f->fp);
 				/* redraw row: counts could be changed */
-				pane_row_draw(p, p->pos);
+				updatesidebar(onlynew);
 				updatetitle();
 			} else if (selpane == PaneItems && panes[PaneItems].nrows) {
 				p = &panes[PaneItems];
 				row = pane_row_get(p, p->pos);
 				item = (struct item *)row->data;
+				markread(p, p->pos, p->pos, 1);
 				plumb(plumber, item->fields[FieldLink]);
 			}
 			break;
@@ -1706,13 +1916,53 @@ nextpage:
 				switch (ch) {
 				case 'y': pipeitem("cut -f 3 | xclip -r", item, 0); break;
 				case 'E': pipeitem("cut -f 8 | xclip -r", item, 0); break;
-				default:  pipeitem(piper, item, 1); break;
+				default:
+					markread(p, p->pos, p->pos, 1);
+					pipeitem(piper, item, 1);
+					break;
 				}
 			}
 			break;
+		case 'f': /* mark all read */
+		case 'F': /* mark all unread */
+			if (panes[PaneItems].nrows) {
+				p = &panes[PaneItems];
+				markread(p, 0, p->nrows - 1, ch == 'f');
+			}
+			break;
+		case 'r': /* mark item as read */
+		case 'u': /* mark item as unread */
+			if (selpane == PaneItems && panes[PaneItems].nrows) {
+				p = &panes[PaneItems];
+				markread(p, p->pos, p->pos, ch == 'r');
+			}
+			break;
 		case 4: /* EOT */
 		case 'q': goto end;
 		}
+event:
+		if (ch == EOF)
+			goto end;
+		else if (ch == -3 && sigstate == 0)
+			continue; /* just a time-out, nothing to do */
+
+		/* handle last signal */
+		switch (sigstate) {
+		case SIGHUP:
+			feeds_reloadall();
+			sigstate = 0;
+			break;
+		case SIGINT:
+		case SIGTERM:
+			cleanup();
+			_exit(128 + sigstate);
+		case SIGWINCH:
+			resizewin();
+			updategeom();
+			sigstate = 0;
+			break;
+		}
+
 		draw();
 	}
 end:
